﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Common;

namespace HeuristicLab.Algorithms.Bandits.BanditPolicies {
  public class UCBNormalPolicy : IBanditPolicy {

    public int SelectAction(Random random, IEnumerable<IBanditPolicyActionInfo> actionInfos) {
      var myActionInfos = actionInfos.OfType<MeanAndVariancePolicyActionInfo>();
      int totalTries = myActionInfos.Sum(a => a.Tries);
      double bestQ = double.NegativeInfinity;
      int aIdx = -1;
      var bestActions = new List<int>();
      foreach (var aInfo in myActionInfos) {
        aIdx++;
        double q;
        if (totalTries <= 1 || aInfo.Tries <= 1 || aInfo.Tries <= Math.Ceiling(8 * Math.Log(totalTries))) {
          q = double.PositiveInfinity;
        } else {
          var tries = aInfo.Tries;
          var avgReward = aInfo.AvgReward;
          var rewardVariance = aInfo.RewardVariance;
          var estVariance = 16.0 * rewardVariance * (Math.Log(totalTries - 1) / tries);
          q = avgReward + Math.Sqrt(estVariance);
        }
        if (q > bestQ) {
          bestQ = q;
          bestActions.Clear();
          bestActions.Add(aIdx);
        } else if (q.IsAlmost(bestQ)) {
          bestActions.Add(aIdx);
        }
      }
      Debug.Assert(bestActions.Any());
      return bestActions.SelectRandom(random);
    }

    public IBanditPolicyActionInfo CreateActionInfo() {
      return new MeanAndVariancePolicyActionInfo();
    }

    public override string ToString() {
      return "UCBNormalPolicy";
    }
  }
}
